module diagnostics
  implicit none
  save
  private

  public :: mix_ratio_to_spec_hum, temp_from_theta_p, surface_pres, sea_pres, &
            lifted_index, mean_rh
  public :: Grav

  ! Some meteorological constants
  real, parameter :: RGas = 287.04, Grav = 9.81, SclHt = RGas * 256. / Grav,  &
                     SpecHeatPresDry = 1004, Kappa = RGas / SpecHeatPresDry,  &
                     Gamma = .0065, LatHeatVap = 2.5e6, Eps = .622

  interface mix_ratio_to_spec_hum
     module procedure mrtsh3d
     module procedure mrtsh4d
  end interface

  interface temp_from_theta_p
     module procedure tftp0d
     module procedure tftp2d
     module procedure tftp3d
  end interface

contains  ! Public Functions ==================================================

  ! Both mrtsh3d and mrtsh4d convert mixing ratio (kg/kg) to specific
  ! humidity (kg/kg).
  function mrtsh3d(w) result (q)
    real, dimension(:,:,:), intent(in)             :: w
    real, dimension(size(w,1),size(w,2),size(w,3)) :: q

    q = w / (1 + w)
  end function mrtsh3d
    
  ! ===========================================================================

  function mrtsh4d(w) result (q)
    real, dimension(:,:,:,:), intent(in)                     :: w
    real, dimension(size(w,1),size(w,2),size(w,3),size(w,4)) :: q
    
    q = w / (1 + w)
  end function mrtsh4d

  ! ===========================================================================

  ! tftp0d, tftp2d, and tftp3d calculate the temperature (K) given the
  ! potential temperature (th; K) and pressure (p; Pa or hPa depending on si
  ! flag).
  function tftp0d(th, p, si) result (t)
    real,    intent(in)           :: th, p
    logical, intent(in), optional :: si
    real                          :: t

    real    :: p0
    
    ! Set appropriate p0
    p0 = 100000.  ! Default is SI units
    if (present(si)) then
       if (.not. si) p0 = 1000.
    end if
    
    ! Calculate temperature
    t = th * (p / p0) ** Kappa
  end function tftp0d

  ! ===========================================================================

  function tftp2d(th, p, si) result (t)
    real, dimension(:,:), intent(in)           :: th, p
    logical,              intent(in), optional :: si
    real, dimension(size(p,1),size(p,2))       :: t

    integer, parameter :: NumDims = 2
    integer :: i
    real    :: p0
    logical :: ok

    ! Make sure arrays match up
    ok = .true.
    do i = 1, NumDims
       if (size(th,i) /= size(p,i)) ok = .false.
    end do
    if (.not. ok) stop "Array extent mismatch in tftp2d!"
    
    ! Set appropriate p0
    p0 = 100000.  ! Default is SI units
    if (present(si)) then
       if (.not. si) p0 = 1000.
    end if
    
    ! Calculate temperature
    t = th * (p / p0) ** Kappa
  end function tftp2d
       
  ! ===========================================================================

  function tftp3d(th, p, si) result (t)
    real, dimension(:,:,:), intent(in)             :: th, p
    logical,                intent(in), optional   :: si
    real, dimension(size(p,1),size(p,2),size(p,3)) :: t
    
    integer, parameter :: NumDims = 3
    integer :: i
    real    :: p0
    logical :: ok
    
    ! Make sure arrays match up
    ok = .true.
    do i = 1, NumDims
       if (size(th,i) /= size(p,i)) ok = .false.
    end do
    if (.not. ok) stop "Array extent mismatch in tftp3d!"
    
    ! Set appropriate p0
    p0 = 100000.  ! Default is SI units
    if (present(si)) then
       if (.not. si) p0 = 1000.
    end if

    ! Calculate temperature
    t = th * (p / p0) ** Kappa
  end function tftp3d

  ! ===========================================================================

  ! surface_pres calculates the surface pressure (Pa) given the following data:
  ! ht-   Geopotential height (m) at lowest two w (full) levels  (x,y,z,t)
  ! p1-   Pressure (Pa) at lowest mass (half) level              (x,y,t)
  ! q1-   Specific humidity (kg/kg) at lowest mass level         (x,y,t)
  ! t1-   Temperature (K) at lowest mass level                   (x,y,t)
  ! sfcq- Specific humidity (kq/kq) at surface/2-m above ground  (x,y,t)
  ! It is assumed that the array extents match appropriately.
  function surface_pres(ht, p1, q1, t1, sfcq)
    real, dimension(:,:,:,:), intent(in)              :: ht
    real, dimension(:,:,:), intent(in)                :: p1, q1, t1, sfcq
    real, dimension(size(p1,1),size(p1,2),size(p1,3)) :: surface_pres

    real, dimension(size(t1,1),size(t1,2),size(t1,3)) :: tvirt

    tvirt = (1 + .608 * .5 * (sfcq + q1)) * t1

    surface_pres = p1 * exp((Grav / (RGas * tvirt))  &
         * .5 * (ht(:,:,2,:) - ht(:,:,1,:)))
  end function surface_pres

  ! ===========================================================================

  ! sea_pres is taken from the WRF NCL tools.  The inputs are:
  ! zs-    Geopotential height (m) on w (full) levels       (x,y,z,t)
  ! theta- Potential temperature (K) on mass (half) levels  (x,y,z,t)
  ! p-     Pressure (Pa) on mass levels                     (x,y,z,t)
  ! q-     Specific humidity (kg/kg) on mass levels         (x,y,z,t)
  ! It is assumed that the array extents match appropriately.
  function sea_pres(zs, theta, p, q)
    ! Estimate sea level pressure.
    real, dimension(:,:,:,:), intent(in)           :: zs, theta, p, q
    real, dimension(size(p,1),size(p,2),size(p,4)) :: sea_pres

    real,    dimension(size(p,1),size(p,2),size(p,3),size(p,4)) :: t, z
    real,    dimension(size(p,1),size(p,2)) :: tSurf, tSeaLevel
    integer, dimension(size(p,1),size(p,2)) :: level
    
    ! Specific constants for assumptions made in this routine:
    real, parameter :: Tc = 273.16+17.5, PConst = 10000
    
    !  Local variables:
    real    :: plo, phi, tlo, thi, zlo, zhi, pAtPConst, tAtPConst, zAtPConst
    real    :: zHalfLowest
    integer :: nx, ny, nz, nt
    integer :: i, j, k, klo, khi, l
    logical :: l1, l2, l3, found
    
    nx = size(p,1)
    ny = size(p,2)
    nz = size(p,3)
    nt = size(p,4)
    
    ! Convert potential temperature to temperature
    t = theta * (p / 100000.) ** Kappa
    
    ! Unstagger z
    do k = 1, nz
       z(:,:,k,:) = .5 * (zs(:,:,k,:) + zs(:,:,k+1,:))
    end do
    
    ! Grand loop over all times
    do l = 1, nt
       ! Find least zeta level that is PConst Pa above the surface.  We later
       ! use this level to extrapolate a surface pressure and temperature,
       ! which is supposed to reduce the effect of the diurnal heating cycle in
       ! the pressure field.
       
       do j = 1 , ny
          do i = 1 , nx
             level(i,j) = -1
             
             k = 1
             found = .false.
             do while( (.not. found) .and. (k <= nz))
                if ( p(i,j,k,l) < p(i,j,1,l)-PConst ) then
                   level(i,j) = k
                   found = .true.
                end if
                k = k+1
             end do
             
             if ( level(i,j) == -1 ) then
                print '(A,I4,A)','Troubles finding level ',  &
                     nint(PConst)/100,' above ground.'
                print '(A,I4,A,I4,A)',  &
                     'Problems first occur at (',i,',',j,')'
                print '(A,F6.1,A)',  &
                     'Surface pressure = ',p(i,j,1,l)/100,' hPa.'
                stop 'Error_in_finding_100_hPa_up'
             end if
                          
          end do
       end do
       
       ! Get temperature PConst Pa above surface.  Use this to extrapolate 
       ! the temperature at the surface and down to sea level.
       
       do j = 1 , ny
          do i = 1 , nx
             
             klo = max ( level(i,j) - 1 , 1      )
             khi = min ( klo + 1        , nz - 1 )
             
             if ( klo == khi ) then
                print '(A)','Trapping levels are weird.'
                print '(A,I3,A,I3,A)','klo = ',klo,', khi = ',khi,  &
                     ': and they should not be equal.'
                stop 'Error_trapping_levels'
             end if
             
             plo = p(i,j,klo,l)
             phi = p(i,j,khi,l)
             tlo = t(i,j,klo,l) * (1. + 0.608 * q(i,j,klo,l) )
             thi = t(i,j,khi,l) * (1. + 0.608 * q(i,j,khi,l) )
             zlo = z(i,j,klo,l)         
             zhi = z(i,j,khi,l)
             
             pAtPConst = p(i,j,1,l) - PConst
             TAtPConst = thi - (thi-tlo) * log(pAtPConst/phi) * log(plo/phi)
             zAtPConst = zhi - (zhi-zlo) * log(pAtPConst/phi) * log(plo/phi)
             
             tSurf(i,j) = tAtPConst * (p(i,j,1,l)/pAtPConst)**(Gamma*RGas/Grav)
             tSeaLevel(i,j) = tAtPConst + Gamma * zAtPConst
             
          end do
       end do

       ! If we follow a traditional computation, there is a correction to the
       ! sea level temperature if both the surface and sea level temnperatures
       ! are *too* hot.
       
       do j = 1 , ny
          do i = 1 , nx
             l1 = tSeaLevel(i,j) <  Tc 
             l2 = tSurf    (i,j) <= Tc
             l3 = .not. l1
             if ( l2 .and. l3 ) then
                tSeaLevel(i,j) = Tc
             else
                tSeaLevel(i,j) = Tc - 0.005 * (tSurf(i,j) - Tc)**2
             end if
          end do
       end do
       
       ! The grand finale: ta da!
       
       do j = 1 , ny
          do i = 1 , nx
             zHalfLowest = z(i,j,1,l)
             sea_pres(i,j,l) = p(i,j,1,l) *  &
                  exp((2.*Grav*zHalfLowest)/(RGas*(tSeaLevel(i,j)+tSurf(i,j))))
          end do
       end do
    end do
  end function sea_pres

  ! ===========================================================================

  ! lifted_index calculates the lifted index (K) based on the supplied profiles
  ! of potential temperature (theta; K), pressure (p; hPa), and mixing ratio
  ! (w; kg/kg).  Array indices are ordered from the ground up.  The lowest 4
  ! model levels are mixed to produce the hypothetical parcel, which is then
  ! lifted adiabatically to saturation, and pseudoadiabatically thereafter.
  function lifted_index(theta, p, w) result (li)
    real, dimension(:), intent(in) :: theta, p, w
    real                           :: li

    real, dimension(2) :: pBounds, thBounds
    real               :: dens, thMix, wMix, sum, tk, pLCL, thes, tParcel, tEnv
    integer            :: nz, k
    
    nz = size(theta)
    
    ! Check that array extents match
    if (nz /= size(p) .or. nz /= size(w))  &
         stop "Array mismatch in lifted_index."
    
    ! Set lifted index to 0 if we don't have enough data
    if (nz < 4 .or. p(1) < 500.) then
       li = 0
    else
       ! Determine temperature of pseudoadiabatically lifted parcel at 500 mb
       ! Step 1: Mix the lowest 4 model levels
       thMix = 0; wMix = 0; sum = 0
       do k = 1, 4
          dens = density(theta(k), p(k))  ! Effect of water vapor ignored here
          thMix = thMix + dens * theta(k)
          wMix = wMix + dens * w(k)
          sum = sum + dens
       end do
       thMix = thMix / sum
       wMix = wMix / sum

       ! Step 2: Lift parcel to LCL
       tk = temp_from_theta_p(thMix, p(1), si=.false.)
       ! This equation is derived from Curry & Webster (1999, p. 174)
       pLCL = 1000 * (temp_lcl(tk, rh(wMix, p(1), tk)) / thMix)**(1./Kappa)
       
       ! Step 3: Lift parcel to 500 mb
       thes = theta_es(thMix, pLCL)
       tParcel = temp_from_theta_p(theta_es_to_theta(thes, 500.), 500.,  &
            si=.false.)

       ! Get environmental temperature at 500 mb
       call find_bounds(p, theta, 500., pBounds, thBounds)
       tEnv = temp_from_theta_p(lin_int(pBounds(1), pBounds(2), thBounds(1),  &
            thBounds(2), 500.), 500., si=.false.)
       
       li = tEnv - tParcel
    end if
  end function lifted_index

  ! ===========================================================================

  ! mean_rh computes the mean relative humidity (%) between pBot hPa and pTop
  ! hPa based on the supplied profiles of potential temperature (theta; K),
  ! pressure (p; hPa), and mixing ratio (w; kg/kg).  Also provided is the
  ! surface pressure (pSfc; hPa).  Array indices are ordered from the ground
  ! up.  The mean is weighted by the thickness of each layer.
  function mean_rh(theta, p, w, pSfc, pBot, pTop) result (mrh)
    real, dimension(:), intent(in) :: theta, p, w
    real,               intent(in) :: pSfc, pBot, pTop
    real                           :: mrh

    real, dimension(0:size(theta)-1) :: pMid
    real                             :: pThick, relHum, totPresThick, totRH
    integer                          :: nz, k

    nz = size(theta)

    ! Check that array extents match
    if (nz /= size(p) .or. nz /= size(w)) stop "Array mismatch in mean_rh."
    
    ! Calculate pressure at midpoints
    pMid(0) = pSfc
    do k = 1, nz-1
       pMid(k) = .5 * (p(k) + p(k+1))
    end do

    ! Calculate mean RH
    totPresThick = 0; totRH = 0
    do k = 1, nz-1
       if (pMid(k) > pBot) cycle
       if (pMid(k-1) < pTop) exit

       ! Calculate layer thickness
       if (pMid(k-1) > pBot) then
          pThick = pBot - pMid(k)
       else if (pMid(k) < pTop) then
          pThick = pMid(k-1) - pTop
       else
          pThick = pMid(k-1) - pMid(k)
       end if

       relHum = rh(w(k), p(k), temp_from_theta_p(theta(k), p(k), si=.false.))
       totRH = totRH + pThick * relHum
       totPresThick = totPresThick + pThick
    end do
    
    if (totPresThick > .01) then
       mrh = 100 * totRH / totPresThick
    else
       mrh = 0
    end if
  end function mean_rh

  ! ===========================================================================
  ! Private functions (called only from this module) go below this line.
  ! ===========================================================================

  ! density uses the ideal gas law to compute the density of dry air given its
  ! potential temperature (theta; K) and pressure (p; hPa)
  real function density(theta, p)
    real, intent(in) :: theta, p
    
    real :: t
    
    t = temp_from_theta_p(theta, p, si=.false.)
    density = 100 * p / (rgas * t)  ! * 100 since p is in hPa
  end function density

  ! ===========================================================================

  ! theta_es computes the saturated equivalent potential temperature given
  ! potential temperature (theta; K) and pressure (p; hPa)
  real function theta_es(theta, p)
    real, intent(in) :: theta, p

    real :: tmpk, es, ws

    tmpk = temp_from_theta_p(theta, p, si=.false.)
    es = sat_vap_pres(tmpk)
    ws = mixr(es, p)

    theta_es = theta * exp(LatHeatVap * ws / (SpecHeatPresDry * tmpk))
  end function theta_es

  ! ===========================================================================

  ! theta_es_to_theta computes the potential temperature given the saturated
  ! equivalent potential temperature (thes; K) and the pressure (p; hPa),
  ! using the secant method.
  function theta_es_to_theta(thes, p) result(theta)
    real, intent(in) :: thes, p
    real             :: theta

    real,    parameter :: Delta = 1e-4, Epsilon = 1e-4
    integer, parameter :: MaxIter = 200

    real    :: a, b, fa, fb, s
    integer :: k
    
    ! Make initial guess
    a = thes - 10
    b = thes - 20
    fa = f(a)
    fb = f(b)
    
    ! Here's the secant method, at the end of which a is the temperature.
    do k = 2, MaxIter
       if (abs(fa) > abs(fb)) then
          call swap(a, b)
          call swap(fa, fb)
       end if
       s = (b - a) / (fb - fa)
       b = a
       fb = fa
       a = a - fa * s
       fa = f(a)
       if (abs(fa) < Epsilon .or. abs(b-a) < Delta) exit
    end do
    
    theta = a * (1000. / p)**Kappa
    
  contains
    
    ! This function has a zero when the temperature satisfies the equation
    ! for saturated equivalent potential temperature.
    real function f(t)
      real, intent(in) :: t
      
      real :: es, ws
      
      es = sat_vap_pres(t)
      ws = mixr(es, p)
      
      ! Compare f to function theta_es
      f = thes - t * (1000./p)**kappa *  &
           exp(LatHeatVap * ws / (SpecHeatPresDry * t))
    end function f
  end function theta_es_to_theta
  
  ! ===========================================================================

  ! mixr calculates the mixing ratio (kg/kg) given the vapor pressure (e) and
  ! pressure (p). e and p must have the same units.
  real function mixr(e, p)
    real, intent(in) :: e, p

    mixr = Eps * e / (p - e)
  end function mixr

  ! ===========================================================================

  ! vap_pres calculates the vapor pressure given the mixing ratio (w; kg/kg)
  ! and pressure (p).  The vapor pressure will be in the same units as p.
  real function vap_pres(w, p)
    real, intent(in) :: w, p
    
    vap_pres = w * p / (Eps + w)
  end function vap_pres

  ! ===========================================================================

  ! sat_vap_pres uses the data from Curry & Webster (1999, p.113) to compute
  ! the saturation vapor pressure (hPa) given temperature (tmpk; K).
  real function sat_vap_pres(tmpk)
    real, intent(in) :: tmpk

    real, dimension(7), parameter :: a = (/ 6.1117675, .443986062,  &
         .143053301e-1, .265027242e-3, .302246994e-5, .203886313e-7,  &
         .638780966e-10 /)
    
    real    :: psum, tdiff
    integer :: i
    
    tdiff = tmpk - 273.15
    psum = 0
    do i = 7, 2, -1
       psum = psum + a(i) * tdiff**(i-1)
    end do
    sat_vap_pres = psum + a(1)
  end function sat_vap_pres

  ! ===========================================================================

  ! temp_lcl calculates the temperature at the LCL given a parcel's temperature
  ! (tmpk; K) and relative humidity (rh; 0<rh<1) using Curry & Webster (1999,
  ! Eq. 6.30).
  real function temp_lcl(tmpk, rh)
    real, intent(in) :: tmpk, rh

    real :: a, b
    
    a = 1. / (tmpk - 55)
    b = log(rh) / 2840.
    temp_lcl = 1. / (a - b) + 55
  end function temp_lcl

  ! ===========================================================================

  ! rh calculates the relative humidity (0<rh<1) given the mixing ratio
  ! (w; kg/kg), pressure (p; hPa), and temperature (tmpk; K).
  real function rh(w, p, tmpk)
    real, intent(in) :: w, p, tmpk

    real :: es, e

    es = sat_vap_pres(tmpk)
    e = vap_pres(w, p)
    
    rh = e/es
  end function rh

  ! ===========================================================================

  ! swap interchanges the values of a and b.
  subroutine swap(a, b)
    real, intent(inout) :: a, b

    real :: temp
    
    temp = a
    a = b
    b = temp
  end subroutine swap

  ! ===========================================================================

  ! Given profiles of variables a (aProf) and b (bProf) and a target value of
  ! variable a (a), find_bounds outputs two adjacent values in aProf that
  ! bound a (aBound) and the associated values of b (bBound).
  ! It is assumed that the values of aProf decrease with increasing index.
  subroutine find_bounds(aProf, bProf, a, aBound, bBound)
    real, dimension(:), intent(in)  :: aProf, bProf
    real,               intent(in)  :: a
    real, dimension(2), intent(out) :: aBound, bBound

    integer :: i, aSize
    logical :: found
    
    ! Make sure input arrays match
    aSize = size(aProf)
    if (aSize /= size(bProf)) then
       print *, "aProf and bProf have different sizes!", aSize, size(bProf)
       stop
    end if

    ! Find bounds
    found = .false.
    do i = 1, aSize-1
       if ((aProf(i) >= a) .and. (aProf(i+1) < a)) then
          found = .true.
          aBound(1) = aProf(i)
          aBound(2) = aProf(i+1)
          bBound(1) = bProf(i)
          bBound(2) = bProf(i+1)
          exit
       end if
    end do

    if (.not. found) then
       print *, "Could not find bounds!"
       print *, aProf, a
       stop
    end if
  end subroutine find_bounds

  ! ===========================================================================

  ! lin_int performs linear interpolation given two data points (x1,y1) and 
  ! (x2,y2) and a known value for x.
  real function lin_int(x1, x2, y1, y2, x)
    real, intent(in) :: x1, x2, y1, y2, x
    
    if (x1 /= x2) then
       lin_int = y1 + (y2-y1)*(x-x1)/(x2-x1)
    else
       lin_int = y1
    end if
  end function lin_int
end module diagnostics
